#!/usr/bin/env python

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


# mozhttpd web server.

import argparse
import os
import socket

import mozhttpd

# directory of this file
here = os.path.dirname(os.path.realpath(__file__))


class WebServers:
    def __init__(self, host, port, docroot, count):
        self.host = host
        self.port = port
        self.docroot = docroot
        self.count = count
        self.servers = []

    def start(self):
        self.stop()
        self.servers = []
        port = self.port
        num_errors = 0
        while len(self.servers) < self.count:
            self.servers.append(
                mozhttpd.MozHttpd(host=self.host, port=port, docroot=self.docroot)
            )
            try:
                self.servers[-1].start()
            except OSError as error:
                if isinstance(error, socket.error):
                    if error.errno == 98:
                        print(f"port {port} is in use.")
                    else:
                        print(f"port {port} error {error}")
                elif isinstance(error, str):
                    print(f"port {port} error {error}")
                self.servers.pop()
                num_errors += 1
            except Exception as error:
                print(f"port {port} error {error}")
                self.servers.pop()
                num_errors += 1

            if num_errors > 15:
                raise Exception("Too many errors in webservers.py")
            port += 1

    def stop(self):
        while len(self.servers) > 0:
            server = self.servers.pop()
            server.stop()


def main():
    parser = argparse.ArgumentParser(
        description="Start mozhttpd servers for use by areweslimyet."
    )

    parser.add_argument(
        "--port",
        type=int,
        default=8001,
        help="Starting port. Defaults to 8001. Web servers will be "
        "created for each port from the starting port to starting port "
        "+ count - 1.",
    )
    parser.add_argument(
        "--count",
        type=int,
        default=100,
        help="Number of web servers to start. Defaults to 100.",
    )
    parser.add_argument(
        "--host",
        type=str,
        default="localhost",
        help="Name of webserver host. Defaults to localhost.",
    )

    args = parser.parse_args()
    web_servers = WebServers(args.host, args.port, "%s/html" % here, args.count)
    web_servers.start()


if __name__ == "__main__":
    main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import os
import sys

AWSY_PATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
if AWSY_PATH not in sys.path:
    sys.path.append(AWSY_PATH)

from awsy.awsy_test_case import AwsyTestCase

# A description of each checkpoint and the root path to it.
CHECKPOINTS = [
    {
        "name": "After tabs open [+30s, forced GC]",
        "path": "memory-report-TabsOpenForceGC-4.json.gz",
        "name_filter": ["web ", "Web Content"],  # We only want the content process
        "median": True,  # We want the median from all content processes
    },
]

# A description of each perfherder suite and the path to its values.
PERF_SUITES = [
    {"name": "Base Content Resident Unique Memory", "node": "resident-unique"},
    {"name": "Base Content Heap Unclassified", "node": "explicit/heap-unclassified"},
    {"name": "Base Content JS", "node": "js-main-runtime/", "alertThreshold": 0.25},
    {"name": "Base Content Explicit", "node": "explicit/"},
]


class TestMemoryUsage(AwsyTestCase):
    """
    Provides a base case test that just loads about:memory and reports the
    memory usage of a single content process.
    """

    def urls(self):
        return self._urls

    def perf_suites(self):
        return PERF_SUITES

    def perf_checkpoints(self):
        return CHECKPOINTS

    def perf_extra_opts(self):
        return self._extra_opts

    def setUp(self):
        AwsyTestCase.setUp(self)
        self.logger.info("setting up!")

        # Override AwsyTestCase value, this is always going to be 1 iteration.
        self._iterations = 1

        # Override "entities" from our configuration.
        #
        # We aim to load a number of about:blank pages exactly matching the
        # number of content processes we can have. After this we should no
        # longer have a preallocated content process (although to be sure we
        # explicitly drop it at the end of the test).
        process_count = self.marionette.get_pref("dom.ipc.processCount")
        self._pages_to_load = process_count
        self._urls = ["about:blank"] * process_count

        if self.marionette.get_pref("fission.autostart"):
            self._extra_opts = ["fission"]
        else:
            self._extra_opts = None

        self.logger.info(
            "areweslimyet run by %d pages, "
            "%d iterations, %d perTabPause, %d settleWaitTime, "
            "%d content processes"
            % (
                self._pages_to_load,
                self._iterations,
                self._perTabPause,
                self._settleWaitTime,
                process_count,
            )
        )
        self.logger.info("done setting up!")

    def tearDown(self):
        self.logger.info("tearing down!")
        AwsyTestCase.tearDown(self)
        self.logger.info("done tearing down!")

    def set_preallocated_process_enabled_state(self, enabled):
        """Sets the pref that controls whether we have a preallocated content
        process to the given value.

        This will cause the preallocated process to be destroyed or created
        as appropriate.
        """
        if enabled:
            self.logger.info("re-enabling preallocated process")
        else:
            self.logger.info("disabling preallocated process")
        self.marionette.set_pref("dom.ipc.processPrelaunch.enabled", enabled)

    def test_open_tabs(self):
        """Marionette test entry that returns an array of checkpoint arrays.

        This will generate a set of checkpoints for each iteration requested.
        Upon successful completion the results will be stored in
        |self.testvars["results"]| and accessible to the test runner via the
        |testvars| object it passed in.
        """
        # setup the results array
        results = [[] for _ in range(self.iterations())]

        def create_checkpoint(name, iteration, minimize=False):
            checkpoint = self.do_memory_report(name, iteration, minimize)
            self.assertIsNotNone(checkpoint, "Checkpoint was recorded")
            results[iteration].append(checkpoint)

        # As long as we force the number of iterations to 1 in setUp() above,
        # we don't need to loop over this work.
        assert self._iterations == 1
        self.open_pages()
        self.set_preallocated_process_enabled_state(False)
        self.settle()
        self.settle()
        create_checkpoint("TabsOpenForceGC", 0, minimize=True)
        self.set_preallocated_process_enabled_state(True)
        # (If we wanted to do something after the preallocated process has been
        # recreated, we should call self.settle() again to wait for it.)

        # TODO(ER): Temporary hack until bug 1121139 lands
        self.logger.info("setting results")
        self.testvars["results"] = results



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import argparse
import logging
import os
import sys

import mozinfo
from mach.decorators import Command, CommandArgument, CommandArgumentGroup
from mozbuild.base import BinaryNotFoundException
from mozbuild.base import MachCommandConditions as conditions


def setup_awsy_argument_parser():
    from marionette_harness.runtests import MarionetteArguments
    from mozlog.structured import commandline

    parser = MarionetteArguments()
    commandline.add_logging_group(parser)

    return parser


from awsy import ITERATIONS, MAX_TABS, PER_TAB_PAUSE, SETTLE_WAIT_TIME


def run_awsy(command_context, tests, binary=None, **kwargs):
    import json

    from marionette_harness.runtests import MarionetteHarness, MarionetteTestRunner
    from mozlog.structured import commandline

    parser = setup_awsy_argument_parser()

    awsy_source_dir = os.path.join(command_context.topsrcdir, "testing", "awsy")
    if not tests:
        if kwargs["base"]:
            filename = "test_base_memory_usage.py"
        else:
            filename = "test_memory_usage.py"
        tests = [os.path.join(awsy_source_dir, "awsy", filename)]

    args = argparse.Namespace(tests=tests)

    args.binary = binary

    if kwargs["quick"]:
        kwargs["entities"] = 3
        kwargs["iterations"] = 1
        kwargs["perTabPause"] = 1
        kwargs["settleWaitTime"] = 1

    runtime_testvars = {}
    for arg in (
        "webRootDir",
        "pageManifest",
        "resultsDir",
        "entities",
        "iterations",
        "perTabPause",
        "settleWaitTime",
        "maxTabs",
        "dmd",
        "tp5",
    ):
        if arg in kwargs and kwargs[arg] is not None:
            runtime_testvars[arg] = kwargs[arg]

    if "webRootDir" not in runtime_testvars:
        awsy_tests_dir = os.path.join(command_context.topobjdir, "_tests", "awsy")
        web_root_dir = os.path.join(awsy_tests_dir, "html")
        runtime_testvars["webRootDir"] = web_root_dir
    else:
        web_root_dir = runtime_testvars["webRootDir"]
        awsy_tests_dir = os.path.dirname(web_root_dir)

    if "resultsDir" not in runtime_testvars:
        runtime_testvars["resultsDir"] = os.path.join(awsy_tests_dir, "results")

    runtime_testvars["bin"] = binary
    runtime_testvars["run_local"] = True

    page_load_test_dir = os.path.join(web_root_dir, "page_load_test")
    if not os.path.isdir(page_load_test_dir):
        os.makedirs(page_load_test_dir)

    if not os.path.isdir(runtime_testvars["resultsDir"]):
        os.makedirs(runtime_testvars["resultsDir"])

    runtime_testvars_path = os.path.join(awsy_tests_dir, "runtime-testvars.json")
    if kwargs["testvars"]:
        kwargs["testvars"].append(runtime_testvars_path)
    else:
        kwargs["testvars"] = [runtime_testvars_path]

    runtime_testvars_file = open(runtime_testvars_path, "w")
    runtime_testvars_file.write(json.dumps(runtime_testvars, indent=2))
    runtime_testvars_file.close()

    manifest_file = os.path.join(awsy_source_dir, "tp5n-pageset.manifest")
    tooltool_args = {
        "args": [
            sys.executable,
            os.path.join(command_context.topsrcdir, "mach"),
            "artifact",
            "toolchain",
            "-v",
            "--tooltool-manifest=%s" % manifest_file,
            "--cache-dir=%s"
            % os.path.join(command_context._mach_context.state_dir, "tooltool-cache"),
        ]
    }
    command_context.run_process(cwd=page_load_test_dir, **tooltool_args)
    tp5nzip = os.path.join(page_load_test_dir, "tp5n.zip")
    tp5nmanifest = os.path.join(page_load_test_dir, "tp5n", "tp5n.manifest")
    if not os.path.exists(tp5nmanifest):
        unzip_args = {"args": ["unzip", "-q", "-o", tp5nzip, "-d", page_load_test_dir]}
        try:
            command_context.run_process(**unzip_args)
        except Exception as exc:
            troubleshoot = ""
            if mozinfo.os == "win":
                troubleshoot = (
                    " Try using --web-root to specify a "
                    "directory closer to the drive root."
                )

            command_context.log(
                logging.ERROR,
                "awsy",
                {"directory": page_load_test_dir, "exception": exc},
                "Failed to unzip `tp5n.zip` into "
                "`{directory}` with `{exception}`." + troubleshoot,
            )
            raise exc

    # If '--preferences' was not specified supply our default set.
    if not kwargs["prefs_files"]:
        kwargs["prefs_files"] = [os.path.join(awsy_source_dir, "conf", "prefs.json")]

    # Setup DMD env vars if necessary.
    if kwargs["dmd"]:
        bin_dir = os.path.dirname(binary)

        if "DMD" not in os.environ:
            os.environ["DMD"] = "1"

        # Work around a startup crash with DMD on windows
        if mozinfo.os == "win":
            kwargs["pref"] = "security.sandbox.content.level:0"
            command_context.log(
                logging.WARNING,
                "awsy",
                {},
                "Forcing 'security.sandbox.content.level' = 0 because DMD is enabled.",
            )
        elif mozinfo.os == "mac":
            # On mac binary is in MacOS and dmd.py is in Resources, ie:
            #   Name.app/Contents/MacOS/libdmd.dylib
            #   Name.app/Contents/Resources/dmd.py
            bin_dir = os.path.join(bin_dir, "../Resources/")

        # Also add the bin dir to the python path so we can use dmd.py
        if bin_dir not in sys.path:
            sys.path.append(bin_dir)

    for k, v in kwargs.items():
        setattr(args, k, v)

    parser.verify_usage(args)

    args.logger = commandline.setup_logging(
        "Are We Slim Yet Tests", args, {"mach": sys.stdout}
    )
    failed = MarionetteHarness(MarionetteTestRunner, args=vars(args)).run()
    if failed > 0:
        return 1
    else:
        return 0


@Command(
    "awsy-test",
    category="testing",
    description="Run Are We Slim Yet (AWSY) memory usage testing using marionette.",
    parser=setup_awsy_argument_parser,
)
@CommandArgumentGroup("AWSY")
@CommandArgument(
    "--web-root",
    group="AWSY",
    action="store",
    type=str,
    dest="webRootDir",
    help="Path to web server root directory. If not specified, "
    "defaults to topobjdir/_tests/awsy/html.",
)
@CommandArgument(
    "--page-manifest",
    group="AWSY",
    action="store",
    type=str,
    dest="pageManifest",
    help="Path to page manifest text file containing a list "
    "of urls to test. The urls must be served from localhost. If not "
    "specified, defaults to page_load_test/tp5n/tp5n.manifest under "
    "the web root.",
)
@CommandArgument(
    "--results",
    group="AWSY",
    action="store",
    type=str,
    dest="resultsDir",
    help="Path to results directory. If not specified, defaults "
    "to the parent directory of the web root.",
)
@CommandArgument(
    "--quick",
    group="AWSY",
    action="store_true",
    dest="quick",
    default=False,
    help="Set --entities=3, --iterations=1, --per-tab-pause=1, "
    "--settle-wait-time=1 for a quick test. Overrides any explicit "
    "argument settings.",
)
@CommandArgument(
    "--entities",
    group="AWSY",
    action="store",
    type=int,
    dest="entities",
    help="Number of urls to load. Defaults to the total number of urls.",
)
@CommandArgument(
    "--max-tabs",
    group="AWSY",
    action="store",
    type=int,
    dest="maxTabs",
    help="Maximum number of tabs to open. Defaults to %s." % MAX_TABS,
)
@CommandArgument(
    "--iterations",
    group="AWSY",
    action="store",
    type=int,
    dest="iterations",
    help="Number of times to run through the test suite. "
    "Defaults to %s." % ITERATIONS,
)
@CommandArgument(
    "--per-tab-pause",
    group="AWSY",
    action="store",
    type=int,
    dest="perTabPause",
    help="Seconds to wait in between opening tabs. Defaults to %s." % PER_TAB_PAUSE,
)
@CommandArgument(
    "--settle-wait-time",
    group="AWSY",
    action="store",
    type=int,
    dest="settleWaitTime",
    help="Seconds to wait for things to settled down. "
    "Defaults to %s." % SETTLE_WAIT_TIME,
)
@CommandArgument(
    "--dmd",
    group="AWSY",
    action="store_true",
    dest="dmd",
    default=False,
    help="Enable DMD during testing. Requires a DMD-enabled build.",
)
@CommandArgument(
    "--tp5",
    group="AWSY",
    action="store_true",
    dest="tp5",
    default=False,
    help="Use the tp5 pageset during testing.",
)
@CommandArgument(
    "--base",
    group="AWSY",
    action="store_true",
    dest="base",
    default=False,
    help="Run base memory usage tests.",
)
def run_awsy_test(command_context, tests, **kwargs):
    """mach awsy-test runs the in-tree version of the Are We Slim Yet
    (AWSY) tests.

    awsy-test is implemented as a marionette test and marionette
    test arguments also apply although they are not necessary
    since reasonable defaults will be chosen.

    The AWSY specific arguments can be found in the Command
    Arguments for AWSY section below.

    awsy-test will automatically download the tp5n.zip talos
    pageset from tooltool and install it under
    topobjdir/_tests/awsy/html. You can specify your own page set
    by specifying --web-root and --page-manifest.

    The results of the test will be placed in the results
    directory specified by the --results argument.

    On Windows, you may experience problems due to path length
    errors when extracting the tp5n.zip file containing the
    test pages or when attempting to write checkpoints to the
    results directory. In that case, you should specify both
    the --web-root and --results arguments pointing to a location
    with a short path. For example:

    --web-root=c:\\\\tmp\\\\html --results=c:\\\\tmp\\\\results

    Note that the double backslashes are required.
    """
    kwargs["logger_name"] = "Awsy Tests"
    if "test_objects" in kwargs:
        tests = []
        for obj in kwargs["test_objects"]:
            tests.append(obj["file_relpath"])
        del kwargs["test_objects"]

    if not kwargs.get("binary") and conditions.is_firefox(command_context):
        try:
            kwargs["binary"] = command_context.get_binary_path("app")
        except BinaryNotFoundException as e:
            command_context.log(
                logging.ERROR, "awsy", {"error": str(e)}, "ERROR: {error}"
            )
            command_context.log(logging.INFO, "awsy", {"help": e.help()}, "{help}")
            return 1
    return run_awsy(command_context, tests, **kwargs)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import os
import sys

import mozinfo
import yaml
from marionette_driver.errors import JavascriptException, ScriptTimeoutException
from mozproxy import get_playback

AWSY_PATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
if AWSY_PATH not in sys.path:
    sys.path.append(AWSY_PATH)

from awsy import process_perf_data, webservers
from awsy.awsy_test_case import AwsyTestCase


class TestMemoryUsage(AwsyTestCase):
    """Provides a test that collects memory usage at various checkpoints:
    - "Start" - Just after startup
    - "StartSettled" - After an additional wait time
    - "TabsOpen" - After opening all provided URLs
    - "TabsOpenSettled" - After an additional wait time
    - "TabsOpenForceGC" - After forcibly invoking garbage collection
    - "TabsClosed" - After closing all tabs
    - "TabsClosedSettled" - After an additional wait time
    - "TabsClosedForceGC" - After forcibly invoking garbage collection
    """

    def urls(self):
        return self._urls

    def perf_suites(self):
        return process_perf_data.PERF_SUITES

    def perf_checkpoints(self):
        return process_perf_data.CHECKPOINTS

    def perf_extra_opts(self):
        return self._extra_opts

    def setupTp5(self):
        urls = None
        default_tp5n_manifest = os.path.join(
            self._webroot_dir, "page_load_test", "tp5n", "tp5n.manifest"
        )
        tp5n_manifest = self.testvars.get("pageManifest", default_tp5n_manifest)
        with open(tp5n_manifest) as fp:
            urls = fp.readlines()
        # pylint --py3k: W1636
        urls = list(map(lambda x: x.replace("localhost", "localhost:{}"), urls))

        # We haven't set self._urls yet, so this value might be zero if
        # 'entities' wasn't specified.
        to_load = self.pages_to_load()
        if not to_load:
            to_load = len(urls)
        self._webservers = webservers.WebServers(
            "localhost", 8001, self._webroot_dir, to_load
        )
        self._webservers.start()
        for url, server in zip(urls, self._webservers.servers):
            self._urls.append(url.strip().format(server.port))

    def setupTp6(self):
        # tp5n stores its manifest in the zip file that gets extracted, tp6
        # doesn't so we just keep one in our project dir for now.
        default_tp6_pages_manifest = os.path.join(AWSY_PATH, "conf", "tp6-pages.yml")
        tp6_pages_manifest = self.testvars.get(
            "pageManifest", default_tp6_pages_manifest
        )
        urls = []
        with open(tp6_pages_manifest) as f:
            d = yaml.safe_load(f)
            for r in d:
                url = r["url"]
                if isinstance(url, list):
                    urls.extend(url)
                else:
                    urls.append(url)

        self._urls = urls

        # Indicate that we're using tp6 in the perf data.
        self._extra_opts = ["tp6"]

        if self.marionette.get_pref("fission.autostart"):
            self._extra_opts.append("fission")

        # Now we setup the mitm proxy with our tp6 pageset.
        tp6_pageset_manifest = os.path.join(AWSY_PATH, "tp6-pageset.manifest")
        config = {
            "playback_tool": "mitmproxy",
            "playback_version": "5.1.1",
            "playback_files": [tp6_pageset_manifest],
            "platform": mozinfo.os,
            "obj_path": self._webroot_dir,
            "binary": self._binary,
            "run_local": self._run_local,
            "app": "firefox",
            "host": "127.0.0.1",
            "ignore_mitmdump_exit_failure": True,
        }

        self._playback = get_playback(config)
        self._playback.start()

        # We need to reload after the mitmproxy cert is installed
        self.marionette.restart(in_app=False, clean=False)

        # Setup WebDriver capabilities that we need
        self.marionette.delete_session()
        caps = {
            "unhandledPromptBehavior": "dismiss",  # Ignore page navigation warnings
        }
        self.marionette.start_session(caps)
        self.marionette.set_context("chrome")

    def setUp(self):
        AwsyTestCase.setUp(self)
        self.logger.info("setting up")
        self._webroot_dir = self.testvars["webRootDir"]
        self._urls = []
        self._extra_opts = None

        if self.testvars.get("tp5", False):
            self.setupTp5()
        else:
            self.setupTp6()

        self.logger.info(
            "areweslimyet run by %d pages, %d iterations,"
            " %d perTabPause, %d settleWaitTime"
            % (
                self._pages_to_load,
                self._iterations,
                self._perTabPause,
                self._settleWaitTime,
            )
        )
        self.logger.info("done setting up!")

    def tearDown(self):
        self.logger.info("tearing down!")

        self.logger.info("tearing down webservers!")

        if self.testvars.get("tp5", False):
            self._webservers.stop()
        else:
            self._playback.stop()

        AwsyTestCase.tearDown(self)

        self.logger.info("done tearing down!")

    def clear_preloaded_browser(self):
        """
        Clears out the preloaded browser.
        """
        self.logger.info("closing preloaded browser")
        script = """
            if (window.NewTabPagePreloading) {
                return NewTabPagePreloading.removePreloadedBrowser(window);
            }
            return "NewTabPagePreloading.removePreloadedBrowser not available";
            """
        try:
            result = self.marionette.execute_script(script, script_timeout=180000)
        except JavascriptException as e:
            self.logger.error("removePreloadedBrowser() JavaScript error: %s" % e)
        except ScriptTimeoutException:
            self.logger.error("removePreloadedBrowser() timed out")
        except Exception:
            self.logger.error(
                "removePreloadedBrowser() Unexpected error: %s" % sys.exc_info()[0]
            )
        else:
            if result:
                self.logger.info(result)

    def test_open_tabs(self):
        """Marionette test entry that returns an array of checkpoint arrays.

        This will generate a set of checkpoints for each iteration requested.
        Upon successful completion the results will be stored in
        |self.testvars["results"]| and accessible to the test runner via the
        |testvars| object it passed in.
        """
        # setup the results array
        results = [[] for _ in range(self.iterations())]

        def create_checkpoint(name, iteration, minimize=False):
            checkpoint = self.do_memory_report(name, iteration, minimize)
            self.assertIsNotNone(checkpoint, "Checkpoint was recorded")
            results[iteration].append(checkpoint)

        # The first iteration gets Start and StartSettled entries before
        # opening tabs
        create_checkpoint("Start", 0)
        self.settle()
        create_checkpoint("StartSettled", 0)

        for itr in range(self.iterations()):
            self.open_pages()

            create_checkpoint("TabsOpen", itr)
            self.settle()
            create_checkpoint("TabsOpenSettled", itr)
            create_checkpoint("TabsOpenForceGC", itr, minimize=True)

            # Close all tabs
            self.reset_state()

            with self.marionette.using_context("content"):
                self.logger.info("navigating to about:blank")
                self.marionette.navigate("about:blank")
                self.logger.info("navigated to about:blank")
            self.signal_user_active()

            # Create checkpoint that may contain retained processes that will
            # be reused.
            create_checkpoint("TabsClosedExtraProcesses", itr)

            # Clear out the retained processes and measure again.
            self.clear_preloaded_browser()

            create_checkpoint("TabsClosed", itr)
            self.settle()
            create_checkpoint("TabsClosedSettled", itr)
            create_checkpoint("TabsClosedForceGC", itr, minimize=True)

        # TODO(ER): Temporary hack until bug 1121139 lands
        self.logger.info("setting results")
        self.testvars["results"] = results



#!/usr/bin/env python

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


# Firefox about:memory log parser.

import argparse
import gzip
import json
from collections import defaultdict

# This value comes from nsIMemoryReporter.idl.
KIND_HEAP = 1


def path_total(data, path):
    """
    Calculates the sum for the given data point path and its children. If
    path does not end with a '/' then only the value for the exact path is
    returned.
    """
    path_totals = defaultdict(int)

    # Bookkeeping for calculating the heap-unclassified measurement.
    explicit_heap = defaultdict(int)
    heap_allocated = defaultdict(int)

    discrete = not path.endswith("/")

    def match(value):
        """
        Helper that performs either an explicit match or a prefix match
        depending on the format of the path passed in.
        """
        if discrete:
            return value == path
        else:
            return value.startswith(path)

    def update_bookkeeping(report):
        """
        Adds the value to the heap total if this an explicit entry that is a
        heap measurement and updates the heap allocated value if necessary.
        """
        if report["kind"] == KIND_HEAP and report["path"].startswith("explicit/"):
            explicit_heap[report["process"]] += report["amount"]
        elif report["path"] == "heap-allocated":
            heap_allocated[report["process"]] = report["amount"]

    def heap_unclassified(process):
        """
        Calculates the heap-unclassified value for the given process. This is
        simply the difference between all values reported as heap allocated
        under the explicit/ tree and the value reported for heap-allocated by
        the allocator.
        """
        # Memory reports should always include heap-allocated. If it's missing
        # just assert.
        assert process in heap_allocated

        unclassified = heap_allocated[process] - explicit_heap[process]

        # Make sure the value is sane. A misbehaving reporter could lead to
        # negative values.
        # This assertion fails on Beta while running TP6, in the Google Docs process.
        # Disable this for now, but only on Beta. See bug 1735556.
        # assert unclassified >= 0, "heap-unclassified was negative: %d" % unclassified

        return unclassified

    needs_bookkeeping = path in ("explicit/", "explicit/heap-unclassified")

    # Process all the reports.
    for report in data["reports"]:
        if needs_bookkeeping:
            update_bookkeeping(report)

        if match(report["path"]):
            path_totals[report["process"]] += report["amount"]

    # Handle special processing for explicit and heap-unclassified.
    if path == "explicit/":
        # If 'explicit/' is requested we need to add the 'explicit/heap-unclassified'
        # node that is generated by about:memory.
        for k, v in explicit_heap.items():
            path_totals[k] += heap_unclassified(k)
    elif path == "explicit/heap-unclassified":
        # If 'explicit/heap-unclassified' is requested we need to calculate the
        # value as it's generated by about:memory, not explicitly reported.
        for k, v in explicit_heap.items():
            path_totals[k] = heap_unclassified(k)

    return path_totals


def calculate_memory_report_values(
    memory_report_path, data_point_path, process_names=None
):
    """
    Opens the given memory report file and calculates the value for the given
    data point.

    :param memory_report_path: Path to the memory report file to parse.
    :param data_point_path: Path of the data point to calculate in the memory
     report, ie: 'explicit/heap-unclassified'.
    :param process_name: Name of processes to limit reports to. ie 'Main'
    """
    try:
        with open(memory_report_path) as f:
            data = json.load(f)
    except ValueError:
        # Check if the file is gzipped.
        with gzip.open(memory_report_path, "rb") as f:
            data = json.load(f)

    totals = path_total(data, data_point_path)

    # If a process name is provided, restricted output to processes matching
    # that name.
    if process_names is not None:
        for k in list(totals.keys()):
            if not any([process_name in k for process_name in process_names]):
                del totals[k]

    return totals


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Extract data points from about:memory reports"
    )
    parser.add_argument("report", action="store", help="Path to a memory report file.")
    parser.add_argument(
        "prefix",
        action="store",
        help="Prefix of data point to measure. "
        "If the prefix does not end in a '/' "
        "then an exact match is made.",
    )
    parser.add_argument(
        "--proc-filter",
        action="store",
        nargs="*",
        default=None,
        help="Process name filter. " "If not provided all processes will be included.",
    )
    parser.add_argument(
        "--mebi",
        action="store_true",
        help="Output values as mebibytes (instead of bytes)" " to match about:memory.",
    )

    args = parser.parse_args()
    totals = calculate_memory_report_values(args.report, args.prefix, args.proc_filter)

    sorted_totals = sorted(totals.items(), key=lambda item: (-item[1], item[0]))
    for k, v in sorted_totals:
        if v:
            print(f"{k}\t"),
    print("")

    bytes_per_mebibyte = 1024.0 * 1024.0
    for k, v in sorted_totals:
        if v:
            if args.mebi:
                print(f"{v / bytes_per_mebibyte:.2f} MiB"),
            else:
                print(f"{v} bytes"),
            print("\t"),
    print("")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# Maximum number of tabs to open
MAX_TABS = 30

# Default amount of seconds to wait in between opening tabs
PER_TAB_PAUSE = 10

# Default amount of seconds to wait for things to be settled down
SETTLE_WAIT_TIME = 30

# Amount of times to run through the test suite
ITERATIONS = 5

__all__ = [
    "MAX_TABS",
    "PER_TAB_PAUSE",
    "SETTLE_WAIT_TIME",
    "ITERATIONS",
    "webservers",
    "process_perf_data",
]



#!/usr/bin/env python
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import glob
import json
import math
import os
import sys

AWSY_PATH = os.path.dirname(os.path.realpath(__file__))
if AWSY_PATH not in sys.path:
    sys.path.append(AWSY_PATH)

import parse_about_memory

# A description of each checkpoint and the root path to it.
CHECKPOINTS = [
    {"name": "Fresh start", "path": "memory-report-Start-0.json.gz"},
    {"name": "Fresh start [+30s]", "path": "memory-report-StartSettled-0.json.gz"},
    {"name": "After tabs open", "path": "memory-report-TabsOpen-4.json.gz"},
    {
        "name": "After tabs open [+30s]",
        "path": "memory-report-TabsOpenSettled-4.json.gz",
    },
    {
        "name": "After tabs open [+30s, forced GC]",
        "path": "memory-report-TabsOpenForceGC-4.json.gz",
    },
    {
        "name": "Tabs closed extra processes",
        "path": "memory-report-TabsClosedExtraProcesses-4.json.gz",
    },
    {"name": "Tabs closed", "path": "memory-report-TabsClosed-4.json.gz"},
    {"name": "Tabs closed [+30s]", "path": "memory-report-TabsClosedSettled-4.json.gz"},
    {
        "name": "Tabs closed [+30s, forced GC]",
        "path": "memory-report-TabsClosedForceGC-4.json.gz",
    },
]

# A description of each perfherder suite and the path to its values.
PERF_SUITES = [
    {"name": "Resident Memory", "node": "resident"},
    {"name": "Explicit Memory", "node": "explicit/"},
    {"name": "Heap Unclassified", "node": "explicit/heap-unclassified"},
    {"name": "JS", "node": "js-main-runtime/"},
    {"name": "Images", "node": "explicit/images/"},
]


def median(values):
    sorted_ = sorted(values)
    # pylint --py3k W1619
    med = int(len(sorted_) / 2)

    if len(sorted_) % 2:
        return sorted_[med]
    # pylint --py3k W1619
    return (sorted_[med - 1] + sorted_[med]) / 2


def update_checkpoint_paths(checkpoint_files, checkpoints):
    """
    Updates checkpoints with memory report file fetched in data_path
    :param checkpoint_files: list of files in data_path
    :param checkpoints: The checkpoints to update the path of.
    """
    target_path = [
        ["Start-", 0],
        ["StartSettled-", 0],
        ["TabsOpen-", -1],
        ["TabsOpenSettled-", -1],
        ["TabsOpenForceGC-", -1],
        ["TabsClosedExtraProcesses-", -1],
        ["TabsClosed-", -1],
        ["TabsClosedSettled-", -1],
        ["TabsClosedForceGC-", -1],
    ]
    for i in range(len(target_path)):
        (name, idx) = target_path[i]
        paths = sorted([x for x in checkpoint_files if name in x])
        if paths:
            indices = [i for i, x in enumerate(checkpoints) if name in x["path"]]
            if indices:
                checkpoints[indices[0]]["path"] = paths[idx]
            else:
                print(f"found files but couldn't find {name}")


def create_suite(
    name, node, data_path, checkpoints=CHECKPOINTS, alertThreshold=None, extra_opts=None
):
    """
    Creates a suite suitable for adding to a perfherder blob. Calculates the
    geometric mean of the checkpoint values and adds that to the suite as
    well.

    :param name: The name of the suite.
    :param node: The path of the data node to extract data from.
    :param data_path: The directory to retrieve data from.
    :param checkpoints: Which checkpoints to include.
    :param alertThreshold: The percentage of change that triggers an alert.
    """
    suite = {"name": name, "subtests": [], "lowerIsBetter": True, "unit": "bytes"}

    if alertThreshold:
        suite["alertThreshold"] = alertThreshold

    opts = []
    if extra_opts:
        opts.extend(extra_opts)

    if "DMD" in os.environ and os.environ["DMD"]:
        opts.append("dmd")

    if len(opts) > 0:
        suite["extraOptions"] = opts

    update_checkpoint_paths(
        glob.glob(os.path.join(data_path, "memory-report*")), checkpoints
    )

    total = 0
    for checkpoint in checkpoints:
        memory_report_path = os.path.join(data_path, checkpoint["path"])

        name_filter = checkpoint.get("name_filter", None)
        if checkpoint.get("median"):
            process = median
        else:
            process = sum

        if node != "resident":
            totals = parse_about_memory.calculate_memory_report_values(
                memory_report_path, node, name_filter
            )
            value = process(totals.values())
        else:
            # For "resident" we really want RSS of the chrome ("Main") process
            # and USS of the child processes. We'll still call it resident
            # for simplicity (it's nice to be able to compare RSS of non-e10s
            # with RSS + USS of e10s).
            totals_rss = parse_about_memory.calculate_memory_report_values(
                memory_report_path, node, ["Main"]
            )
            totals_uss = parse_about_memory.calculate_memory_report_values(
                memory_report_path, "resident-unique"
            )
            value = list(totals_rss.values())[0] + sum(
                [v for k, v in totals_uss.items() if "Main" not in k]
            )

        subtest = {
            "name": checkpoint["name"],
            "value": value,
            "lowerIsBetter": True,
            "unit": "bytes",
        }
        suite["subtests"].append(subtest)
        total += math.log(subtest["value"])

    # Add the geometric mean. For more details on the calculation see:
    #   https://en.wikipedia.org/wiki/Geometric_mean#Relationship_with_arithmetic_mean_of_logarithms
    # pylint --py3k W1619
    suite["value"] = math.exp(total / len(checkpoints))

    return suite


def create_perf_data(
    data_path, perf_suites=PERF_SUITES, checkpoints=CHECKPOINTS, extra_opts=None
):
    """
    Builds up a performance data blob suitable for submitting to perfherder.
    """
    if ("GCOV_PREFIX" in os.environ) or ("JS_CODE_COVERAGE_OUTPUT_DIR" in os.environ):
        print(
            "Code coverage is being collected, performance data will not be gathered."
        )
        return {}

    perf_blob = {"framework": {"name": "awsy"}, "suites": []}

    for suite in perf_suites:
        perf_blob["suites"].append(
            create_suite(
                suite["name"],
                suite["node"],
                data_path,
                checkpoints,
                suite.get("alertThreshold"),
                extra_opts,
            )
        )

    return perf_blob


if __name__ == "__main__":
    args = sys.argv[1:]
    if not args:
        print("Usage: process_perf_data.py data_path")
        sys.exit(1)

    # Determine which revisions we need to process.
    data_path = args[0]
    perf_blob = create_perf_data(data_path)
    print(f"PERFHERDER_DATA: {json.dumps(perf_blob)}")

    sys.exit(0)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

from setuptools import find_packages, setup

PACKAGE_NAME = "awsy"
PACKAGE_VERSION = "0.0.1"

setup(
    name=PACKAGE_NAME,
    version=PACKAGE_VERSION,
    description="AreWeSlimYet",
    long_description="A memory testing framework for Firefox.",
    author="Mozilla Automation and Testing Team",
    author_email="tools@lists.mozilla.org",
    license="MPL 1.1/GPL 2.0/LGPL 2.1",
    packages=find_packages(),
    zip_safe=False,
    install_requires=["marionette_harness", "PyYaml"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Mozilla Public License 1.1 (MPL 1.1)",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



# -*- Mode: python; c-basic-offset: 4; indent-tabs-mode: nil; tab-width: 40 -*-
# vim: set filetype=python:
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import fnmatch
import glob
import gzip
import json
import os
import shutil
import sys
import tempfile
import time

import mozlog.structured
from marionette_driver import Wait
from marionette_driver.errors import JavascriptException, ScriptTimeoutException
from marionette_driver.keys import Keys
from marionette_harness import MarionetteTestCase

AWSY_PATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
if AWSY_PATH not in sys.path:
    sys.path.append(AWSY_PATH)

from awsy import (
    ITERATIONS,
    MAX_TABS,
    PER_TAB_PAUSE,
    SETTLE_WAIT_TIME,
    process_perf_data,
)


class AwsyTestCase(MarionetteTestCase):
    """
    Base test case for AWSY tests.
    """

    def urls(self):
        raise NotImplementedError()

    def perf_suites(self):
        raise NotImplementedError()

    def perf_checkpoints(self):
        raise NotImplementedError()

    def perf_extra_opts(self):
        return None

    def iterations(self):
        return self._iterations

    def pages_to_load(self):
        return self._pages_to_load if self._pages_to_load else len(self.urls())

    def settle(self):
        """
        Pauses for the settle time.
        """
        time.sleep(self._settleWaitTime)

    def setUp(self):
        MarionetteTestCase.setUp(self)

        self.logger = mozlog.structured.structuredlog.get_default_logger()
        self.marionette.set_context("chrome")
        self._resultsDir = self.testvars["resultsDir"]

        self._binary = self.testvars["bin"]
        self._run_local = self.testvars.get("run_local", False)

        # Cleanup our files from previous runs.
        for patt in (
            "memory-report-*.json.gz",
            "perfherder-data.json",
            "dmd-*.json.gz",
        ):
            for f in glob.glob(os.path.join(self._resultsDir, patt)):
                os.unlink(f)

        # Optional testvars.
        self._pages_to_load = self.testvars.get("entities", 0)
        self._iterations = self.testvars.get("iterations", ITERATIONS)
        self._perTabPause = self.testvars.get("perTabPause", PER_TAB_PAUSE)
        self._settleWaitTime = self.testvars.get("settleWaitTime", SETTLE_WAIT_TIME)
        self._maxTabs = self.testvars.get("maxTabs", MAX_TABS)
        self._dmd = self.testvars.get("dmd", False)

        self.logger.info(
            "areweslimyet run by %d pages, %d iterations,"
            " %d perTabPause, %d settleWaitTime"
            % (
                self._pages_to_load,
                self._iterations,
                self._perTabPause,
                self._settleWaitTime,
            )
        )
        self.reset_state()

    def tearDown(self):
        MarionetteTestCase.tearDown(self)

        try:
            self.logger.info("processing data in %s!" % self._resultsDir)
            perf_blob = process_perf_data.create_perf_data(
                self._resultsDir,
                self.perf_suites(),
                self.perf_checkpoints(),
                self.perf_extra_opts(),
            )
            self.logger.info("PERFHERDER_DATA: %s" % json.dumps(perf_blob))

            perf_file = os.path.join(self._resultsDir, "perfherder-data.json")
            with open(perf_file, "w") as fp:
                json.dump(perf_blob, fp, indent=2)
            self.logger.info("Perfherder data written to %s" % perf_file)
        except Exception:
            raise
        finally:
            # Make sure we cleanup and upload any existing files even if there
            # were errors processing the perf data.
            if self._dmd:
                self.cleanup_dmd()

            # copy it to moz upload dir if set
            if "MOZ_UPLOAD_DIR" in os.environ:
                for file in os.listdir(self._resultsDir):
                    file = os.path.join(self._resultsDir, file)
                    if os.path.isfile(file):
                        shutil.copy2(file, os.environ["MOZ_UPLOAD_DIR"])

    def cleanup_dmd(self):
        """
        Handles moving DMD reports from the temp dir to our resultsDir.
        """
        from dmd import fixStackTraces

        # Move DMD files from temp dir to resultsDir.
        tmpdir = tempfile.gettempdir()
        tmp_files = os.listdir(tmpdir)
        for f in fnmatch.filter(tmp_files, "dmd-*.json.gz"):
            f = os.path.join(tmpdir, f)
            # We don't fix stacks on Windows, even though we could, due to the
            # tale of woe in bug 1626272.
            if not sys.platform.startswith("win"):
                self.logger.info("Fixing stacks for %s, this may take a while" % f)
                isZipped = True
                fixStackTraces(f, isZipped, gzip.open)
            shutil.move(f, self._resultsDir)

        # Also attempt to cleanup the unified memory reports.
        for f in fnmatch.filter(tmp_files, "unified-memory-report-*.json.gz"):
            try:
                os.remove(f)
            except OSError:
                self.logger.info("Unable to remove %s" % f)

    def reset_state(self):
        self._pages_loaded = 0

        # Close all tabs except one
        for x in self.marionette.window_handles[1:]:
            self.logger.info("closing window: %s" % x)
            self.marionette.switch_to_window(x)
            self.marionette.close()

        self._tabs = self.marionette.window_handles
        self.marionette.switch_to_window(self._tabs[0])

    def do_memory_report(self, checkpointName, iteration, minimize=False):
        """Creates a memory report for all processes and and returns the
        checkpoint.

        This will block until all reports are retrieved or a timeout occurs.
        Returns the checkpoint or None on error.

        :param checkpointName: The name of the checkpoint.

        :param minimize: If true, minimize memory before getting the report.
        """
        self.logger.info("starting checkpoint %s..." % checkpointName)

        checkpoint_file = "memory-report-%s-%d.json.gz" % (checkpointName, iteration)
        checkpoint_path = os.path.join(self._resultsDir, checkpoint_file)
        # On Windows, replace / with the Windows directory
        # separator \ and escape it to prevent it from being
        # interpreted as an escape character.
        if sys.platform.startswith("win"):
            checkpoint_path = checkpoint_path.replace("\\", "\\\\").replace("/", "\\\\")

        checkpoint_script = r"""
            let [resolve] = arguments;
            let dumper =
            Cc["@mozilla.org/memory-info-dumper;1"].getService(
            Ci.nsIMemoryInfoDumper);
            dumper.dumpMemoryReportsToNamedFile(
                "%s",
                () => resolve("memory report done!"),
                null,
                /* anonymize */ false,
                /* minimize memory usage */ %s);
            """ % (
            checkpoint_path,
            "true" if minimize else "false",
        )

        checkpoint = None
        try:
            finished = self.marionette.execute_async_script(
                checkpoint_script, script_timeout=60000
            )
            if finished:
                checkpoint = checkpoint_path
        except JavascriptException as e:
            self.logger.error("Checkpoint JavaScript error: %s" % e)
        except ScriptTimeoutException:
            self.logger.error("Memory report timed out")
        except Exception:
            self.logger.error("Unexpected error: %s" % sys.exc_info()[0])
        else:
            self.logger.info("checkpoint created, stored in %s" % checkpoint_path)

        # Now trigger a DMD report if requested.
        if self._dmd:
            self.do_dmd(checkpointName, iteration)

        return checkpoint

    def do_dmd(self, checkpointName, iteration):
        """
        Triggers DMD reports that are used to help identify sources of
        'heap-unclassified'.

        NB: This will dump DMD reports to the temp dir. Unfortunately it also
        dumps memory reports, but that's all we have to work with right now.
        """
        self.logger.info("Starting %s DMD reports..." % checkpointName)

        ident = "%s-%d" % (checkpointName, iteration)

        # TODO(ER): This actually takes a minimize argument. We could use that
        # rather than have a separate `do_gc` function. Also it generates a
        # memory report so we could combine this with `do_checkpoint`. The main
        # issue would be moving everything out of the temp dir.
        #
        # Generated files have the form:
        #   dmd-<checkpoint>-<iteration>-pid.json.gz, ie:
        #   dmd-TabsOpenForceGC-0-10885.json.gz
        #
        # and for the memory report:
        #   unified-memory-report-<checkpoint>-<iteration>.json.gz
        dmd_script = (
            r"""
            let dumper =
            Cc["@mozilla.org/memory-info-dumper;1"].getService(
            Ci.nsIMemoryInfoDumper);
            dumper.dumpMemoryInfoToTempDir(
                "%s",
                /* anonymize = */ false,
                /* minimize = */ false);
            """
            % ident
        )

        try:
            # This is async and there's no callback so we use the existence
            # of an incomplete memory report to check if it hasn't finished yet.
            self.marionette.execute_script(dmd_script, script_timeout=60000)
            tmpdir = tempfile.gettempdir()
            prefix = "incomplete-unified-memory-report-%s-%d-*" % (
                checkpointName,
                iteration,
            )
            max_wait = 240
            elapsed = 0
            while fnmatch.filter(os.listdir(tmpdir), prefix) and elapsed < max_wait:
                self.logger.info("Waiting for memory report to finish")
                time.sleep(1)
                elapsed += 1

            incomplete = fnmatch.filter(os.listdir(tmpdir), prefix)
            if incomplete:
                # The memory reports never finished.
                self.logger.error("Incomplete memory reports leftover.")
                for f in incomplete:
                    os.remove(os.path.join(tmpdir, f))

        except JavascriptException as e:
            self.logger.error("DMD JavaScript error: %s" % e)
        except ScriptTimeoutException:
            self.logger.error("DMD timed out")
        except Exception:
            self.logger.error("Unexpected error: %s" % sys.exc_info()[0])
        else:
            self.logger.info("DMD started, prefixed with %s" % ident)

    def open_and_focus(self):
        """Opens the next URL in the list and focuses on the tab it is opened in.

        A new tab will be opened if |_maxTabs| has not been exceeded, otherwise
        the URL will be loaded in the next tab.
        """
        page_to_load = self.urls()[self._pages_loaded % len(self.urls())]
        tabs_loaded = len(self._tabs)
        open_tab_script = r"""
            gBrowser.addTab("about:blank", {
                inBackground: false,
                triggeringPrincipal: Services.scriptSecurityManager.getSystemPrincipal(),
            });
        """

        if tabs_loaded < self._maxTabs and tabs_loaded <= self._pages_loaded:
            full_tab_list = self.marionette.window_handles

            self.marionette.execute_script(open_tab_script, script_timeout=60000)

            Wait(self.marionette).until(
                lambda mn: len(mn.window_handles) == tabs_loaded + 1,
                message="No new tab has been opened",
            )

            # NB: The tab list isn't sorted, so we do a set diff to determine
            #     which is the new tab
            new_tab_list = self.marionette.window_handles
            new_tabs = list(set(new_tab_list) - set(full_tab_list))

            self._tabs.append(new_tabs[0])
            tabs_loaded += 1

        tab_idx = self._pages_loaded % self._maxTabs

        tab = self._tabs[tab_idx]

        # Tell marionette which tab we're on
        # NB: As a work-around for an e10s marionette bug, only select the tab
        #     if we're really switching tabs.
        if tabs_loaded > 1:
            self.logger.info("switching to tab")
            self.marionette.switch_to_window(tab)
            self.logger.info("switched to tab")

        with self.marionette.using_context("content"):
            self.logger.info("loading %s" % page_to_load)
            self.marionette.navigate(page_to_load)
            self.logger.info("loaded!")

        # The tab handle can change after actually loading content
        # First build a set up w/o the current tab
        old_tabs = set(self._tabs)
        old_tabs.remove(tab)
        # Perform a set diff to get the (possibly) new handle
        new_tabs = set(self.marionette.window_handles) - old_tabs
        # Update the tab list at the current index to preserve the tab
        # ordering
        if new_tabs:
            self._tabs[tab_idx] = list(new_tabs)[0]

        # give the page time to settle
        time.sleep(self._perTabPause)

        self._pages_loaded += 1

    def signal_user_active(self):
        """Signal to the browser that the user is active.

        Normally when being driven by marionette the browser thinks the
        user is inactive the whole time because user activity is
        detected by looking at key and mouse events.

        This would be a problem for this test because user inactivity is
        used to schedule some GCs (in particular shrinking GCs), so it
        would make this unrepresentative of real use.

        Instead we manually cause some inconsequential activity (a press
        and release of the shift key) to make the browser think the user
        is active.  Then when we sleep to allow things to settle the
        browser will see the user as becoming inactive and trigger
        appropriate GCs, as would have happened in real use.
        """
        try:
            action = self.marionette.actions.sequence("key", "keyboard_id")
            action.key_down(Keys.SHIFT)
            action.key_up(Keys.SHIFT)
            action.perform()
        finally:
            self.marionette.actions.release()

    def open_pages(self):
        """
        Opens all pages with our given configuration.
        """
        for _ in range(self.pages_to_load()):
            self.open_and_focus()
            self.signal_user_active()
